/******************************************************************************
 *
 * Copyright(c) 2003 - 2011 Intel Corporation. All rights reserved.
 *
 * Portions of this file are derived from the ipw3945 project, as well
 * as portions of the ieee80211 subsystem header files.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA
 *
 * The full GNU General Public License is included in this distribution in the
 * file called LICENSE.
 *
 * Contact Information:
 *  Intel Linux Wireless <ilw@linux.intel.com>
 * Intel Corporation, 5200 N.E. Elam Young Parkway, Hillsboro, OR 97124-6497
 *
 *****************************************************************************/

#include <net mac80211.h="">

#include "iwl-dev.h"
#include "iwl-core.h"
#include "iwl-agn.h"
#include "iwl-trans.h"

/* priv-&gt;shrd-&gt;sta_lock must be held */
static int iwl_sta_ucode_activate(struct iwl_priv *priv, u8 sta_id)
{
	if (sta_id &gt;= IWLAGN_STATION_COUNT) {
		IWL_ERR(priv, "invalid sta_id %u", sta_id);
		return -EINVAL;
	}
	if (!(priv-&gt;stations[sta_id].used &amp; IWL_STA_DRIVER_ACTIVE))
		IWL_ERR(priv, "ACTIVATE a non DRIVER active station id %u "
			"addr %pM\n",
			sta_id, priv-&gt;stations[sta_id].sta.sta.addr);

	if (priv-&gt;stations[sta_id].used &amp; IWL_STA_UCODE_ACTIVE) {
		IWL_DEBUG_ASSOC(priv,
				"STA id %u addr %pM already present in uCode "
				"(according to driver)\n",
				sta_id, priv-&gt;stations[sta_id].sta.sta.addr);
	} else {
		priv-&gt;stations[sta_id].used |= IWL_STA_UCODE_ACTIVE;
		IWL_DEBUG_ASSOC(priv, "Added STA id %u addr %pM to uCode\n",
				sta_id, priv-&gt;stations[sta_id].sta.sta.addr);
	}
	return 0;
}

static int iwl_process_add_sta_resp(struct iwl_priv *priv,
				    struct iwl_addsta_cmd *addsta,
				    struct iwl_rx_packet *pkt)
{
	u8 sta_id = addsta-&gt;sta.sta_id;
	unsigned long flags;
	int ret = -EIO;

	if (pkt-&gt;hdr.flags &amp; IWL_CMD_FAILED_MSK) {
		IWL_ERR(priv, "Bad return from REPLY_ADD_STA (0x%08X)\n",
			pkt-&gt;hdr.flags);
		return ret;
	}

	IWL_DEBUG_INFO(priv, "Processing response for adding station %u\n",
		       sta_id);

	spin_lock_irqsave(&amp;priv-&gt;shrd-&gt;sta_lock, flags);

	switch (pkt-&gt;u.add_sta.status) {
	case ADD_STA_SUCCESS_MSK:
		IWL_DEBUG_INFO(priv, "REPLY_ADD_STA PASSED\n");
		ret = iwl_sta_ucode_activate(priv, sta_id);
		break;
	case ADD_STA_NO_ROOM_IN_TABLE:
		IWL_ERR(priv, "Adding station %d failed, no room in table.\n",
			sta_id);
		break;
	case ADD_STA_NO_BLOCK_ACK_RESOURCE:
		IWL_ERR(priv, "Adding station %d failed, no block ack "
			"resource.\n", sta_id);
		break;
	case ADD_STA_MODIFY_NON_EXIST_STA:
		IWL_ERR(priv, "Attempting to modify non-existing station %d\n",
			sta_id);
		break;
	default:
		IWL_DEBUG_ASSOC(priv, "Received REPLY_ADD_STA:(0x%08X)\n",
				pkt-&gt;u.add_sta.status);
		break;
	}

	IWL_DEBUG_INFO(priv, "%s station id %u addr %pM\n",
		       priv-&gt;stations[sta_id].sta.mode ==
		       STA_CONTROL_MODIFY_MSK ?  "Modified" : "Added",
		       sta_id, priv-&gt;stations[sta_id].sta.sta.addr);

	/*
	 * XXX: The MAC address in the command buffer is often changed from
	 * the original sent to the device. That is, the MAC address
	 * written to the command buffer often is not the same MAC address
	 * read from the command buffer when the command returns. This
	 * issue has not yet been resolved and this debugging is left to
	 * observe the problem.
	 */
	IWL_DEBUG_INFO(priv, "%s station according to cmd buffer %pM\n",
		       priv-&gt;stations[sta_id].sta.mode ==
		       STA_CONTROL_MODIFY_MSK ? "Modified" : "Added",
		       addsta-&gt;sta.addr);
	spin_unlock_irqrestore(&amp;priv-&gt;shrd-&gt;sta_lock, flags);

	return ret;
}

int iwl_add_sta_callback(struct iwl_priv *priv, struct iwl_rx_mem_buffer *rxb,
			       struct iwl_device_cmd *cmd)
{
	struct iwl_rx_packet *pkt = rxb_addr(rxb);
	struct iwl_addsta_cmd *addsta =
		(struct iwl_addsta_cmd *) cmd-&gt;payload;

	return iwl_process_add_sta_resp(priv, addsta, pkt);
}

int iwl_send_add_sta(struct iwl_priv *priv,
		     struct iwl_addsta_cmd *sta, u8 flags)
{
	int ret = 0;
	struct iwl_host_cmd cmd = {
		.id = REPLY_ADD_STA,
		.flags = flags,
		.data = { sta, },
		.len = { sizeof(*sta), },
	};
	u8 sta_id __maybe_unused = sta-&gt;sta.sta_id;

	IWL_DEBUG_INFO(priv, "Adding sta %u (%pM) %ssynchronously\n",
		       sta_id, sta-&gt;sta.addr, flags &amp; CMD_ASYNC ?  "a" : "");

	if (!(flags &amp; CMD_ASYNC)) {
		cmd.flags |= CMD_WANT_SKB;
		might_sleep();
	}

	ret = iwl_trans_send_cmd(trans(priv), &amp;cmd);

	if (ret || (flags &amp; CMD_ASYNC))
		return ret;
	/*else the command was successfully sent in SYNC mode, need to free
	 * the reply page */

	iwl_free_pages(priv-&gt;shrd, cmd.reply_page);

	if (cmd.handler_status)
		IWL_ERR(priv, "%s - error in the CMD response %d", __func__,
			cmd.handler_status);

	return cmd.handler_status;
}

static void iwl_set_ht_add_station(struct iwl_priv *priv, u8 index,
				   struct ieee80211_sta *sta,
				   struct iwl_rxon_context *ctx)
{
	struct ieee80211_sta_ht_cap *sta_ht_inf = &amp;sta-&gt;ht_cap;
	__le32 sta_flags;
	u8 mimo_ps_mode;

	if (!sta || !sta_ht_inf-&gt;ht_supported)
		goto done;

	mimo_ps_mode = (sta_ht_inf-&gt;cap &amp; IEEE80211_HT_CAP_SM_PS) &gt;&gt; 2;
	IWL_DEBUG_ASSOC(priv, "spatial multiplexing power save mode: %s\n",
			(mimo_ps_mode == WLAN_HT_CAP_SM_PS_STATIC) ?
			"static" :
			(mimo_ps_mode == WLAN_HT_CAP_SM_PS_DYNAMIC) ?
			"dynamic" : "disabled");

	sta_flags = priv-&gt;stations[index].sta.station_flags;

	sta_flags &amp;= ~(STA_FLG_RTS_MIMO_PROT_MSK | STA_FLG_MIMO_DIS_MSK);

	switch (mimo_ps_mode) {
	case WLAN_HT_CAP_SM_PS_STATIC:
		sta_flags |= STA_FLG_MIMO_DIS_MSK;
		break;
	case WLAN_HT_CAP_SM_PS_DYNAMIC:
		sta_flags |= STA_FLG_RTS_MIMO_PROT_MSK;
		break;
	case WLAN_HT_CAP_SM_PS_DISABLED:
		break;
	default:
		IWL_WARN(priv, "Invalid MIMO PS mode %d\n", mimo_ps_mode);
		break;
	}

	sta_flags |= cpu_to_le32(
	      (u32)sta_ht_inf-&gt;ampdu_factor &lt;&lt; STA_FLG_MAX_AGG_SIZE_POS);

	sta_flags |= cpu_to_le32(
	      (u32)sta_ht_inf-&gt;ampdu_density &lt;&lt; STA_FLG_AGG_MPDU_DENSITY_POS);

	if (iwl_is_ht40_tx_allowed(priv, ctx, &amp;sta-&gt;ht_cap))
		sta_flags |= STA_FLG_HT40_EN_MSK;
	else
		sta_flags &amp;= ~STA_FLG_HT40_EN_MSK;

	priv-&gt;stations[index].sta.station_flags = sta_flags;
 done:
	return;
}

/**
 * iwl_prep_station - Prepare station information for addition
 *
 * should be called with sta_lock held
 */
u8 iwl_prep_station(struct iwl_priv *priv, struct iwl_rxon_context *ctx,
		    const u8 *addr, bool is_ap, struct ieee80211_sta *sta)
{
	struct iwl_station_entry *station;
	int i;
	u8 sta_id = IWL_INVALID_STATION;

	if (is_ap)
		sta_id = ctx-&gt;ap_sta_id;
	else if (is_broadcast_ether_addr(addr))
		sta_id = ctx-&gt;bcast_sta_id;
	else
		for (i = IWL_STA_ID; i &lt; IWLAGN_STATION_COUNT; i++) {
			if (!compare_ether_addr(priv-&gt;stations[i].sta.sta.addr,
						addr)) {
				sta_id = i;
				break;
			}

			if (!priv-&gt;stations[i].used &amp;&amp;
			    sta_id == IWL_INVALID_STATION)
				sta_id = i;
		}

	/*
	 * These two conditions have the same outcome, but keep them
	 * separate
	 */
	if (unlikely(sta_id == IWL_INVALID_STATION))
		return sta_id;

	/*
	 * uCode is not able to deal with multiple requests to add a
	 * station. Keep track if one is in progress so that we do not send
	 * another.
	 */
	if (priv-&gt;stations[sta_id].used &amp; IWL_STA_UCODE_INPROGRESS) {
		IWL_DEBUG_INFO(priv, "STA %d already in process of being "
			       "added.\n", sta_id);
		return sta_id;
	}

	if ((priv-&gt;stations[sta_id].used &amp; IWL_STA_DRIVER_ACTIVE) &amp;&amp;
	    (priv-&gt;stations[sta_id].used &amp; IWL_STA_UCODE_ACTIVE) &amp;&amp;
	    !compare_ether_addr(priv-&gt;stations[sta_id].sta.sta.addr, addr)) {
		IWL_DEBUG_ASSOC(priv, "STA %d (%pM) already added, not "
				"adding again.\n", sta_id, addr);
		return sta_id;
	}

	station = &amp;priv-&gt;stations[sta_id];
	station-&gt;used = IWL_STA_DRIVER_ACTIVE;
	IWL_DEBUG_ASSOC(priv, "Add STA to driver ID %d: %pM\n",
			sta_id, addr);
	priv-&gt;num_stations++;

	/* Set up the REPLY_ADD_STA command to send to device */
	memset(&amp;station-&gt;sta, 0, sizeof(struct iwl_addsta_cmd));
	memcpy(station-&gt;sta.sta.addr, addr, ETH_ALEN);
	station-&gt;sta.mode = 0;
	station-&gt;sta.sta.sta_id = sta_id;
	station-&gt;sta.station_flags = ctx-&gt;station_flags;
	station-&gt;ctxid = ctx-&gt;ctxid;

	if (sta) {
		struct iwl_station_priv *sta_priv;

		sta_priv = (void *)sta-&gt;drv_priv;
		sta_priv-&gt;ctx = ctx;
	}

	/*
	 * OK to call unconditionally, since local stations (IBSS BSSID
	 * STA and broadcast STA) pass in a NULL sta, and mac80211
	 * doesn't allow HT IBSS.
	 */
	iwl_set_ht_add_station(priv, sta_id, sta, ctx);

	return sta_id;

}

#define STA_WAIT_TIMEOUT (HZ/2)

/**
 * iwl_add_station_common -
 */
int iwl_add_station_common(struct iwl_priv *priv, struct iwl_rxon_context *ctx,
			   const u8 *addr, bool is_ap,
			   struct ieee80211_sta *sta, u8 *sta_id_r)
{
	unsigned long flags_spin;
	int ret = 0;
	u8 sta_id;
	struct iwl_addsta_cmd sta_cmd;

	*sta_id_r = 0;
	spin_lock_irqsave(&amp;priv-&gt;shrd-&gt;sta_lock, flags_spin);
	sta_id = iwl_prep_station(priv, ctx, addr, is_ap, sta);
	if (sta_id == IWL_INVALID_STATION) {
		IWL_ERR(priv, "Unable to prepare station %pM for addition\n",
			addr);
		spin_unlock_irqrestore(&amp;priv-&gt;shrd-&gt;sta_lock, flags_spin);
		return -EINVAL;
	}

	/*
	 * uCode is not able to deal with multiple requests to add a
	 * station. Keep track if one is in progress so that we do not send
	 * another.
	 */
	if (priv-&gt;stations[sta_id].used &amp; IWL_STA_UCODE_INPROGRESS) {
		IWL_DEBUG_INFO(priv, "STA %d already in process of being "
			       "added.\n", sta_id);
		spin_unlock_irqrestore(&amp;priv-&gt;shrd-&gt;sta_lock, flags_spin);
		return -EEXIST;
	}

	if ((priv-&gt;stations[sta_id].used &amp; IWL_STA_DRIVER_ACTIVE) &amp;&amp;
	    (priv-&gt;stations[sta_id].used &amp; IWL_STA_UCODE_ACTIVE)) {
		IWL_DEBUG_ASSOC(priv, "STA %d (%pM) already added, not "
				"adding again.\n", sta_id, addr);
		spin_unlock_irqrestore(&amp;priv-&gt;shrd-&gt;sta_lock, flags_spin);
		return -EEXIST;
	}

	priv-&gt;stations[sta_id].used |= IWL_STA_UCODE_INPROGRESS;
	memcpy(&amp;sta_cmd, &amp;priv-&gt;stations[sta_id].sta,
	       sizeof(struct iwl_addsta_cmd));
	spin_unlock_irqrestore(&amp;priv-&gt;shrd-&gt;sta_lock, flags_spin);

	/* Add station to device's station table */
	ret = iwl_send_add_sta(priv, &amp;sta_cmd, CMD_SYNC);
	if (ret) {
		spin_lock_irqsave(&amp;priv-&gt;shrd-&gt;sta_lock, flags_spin);
		IWL_ERR(priv, "Adding station %pM failed.\n",
			priv-&gt;stations[sta_id].sta.sta.addr);
		priv-&gt;stations[sta_id].used &amp;= ~IWL_STA_DRIVER_ACTIVE;
		priv-&gt;stations[sta_id].used &amp;= ~IWL_STA_UCODE_INPROGRESS;
		spin_unlock_irqrestore(&amp;priv-&gt;shrd-&gt;sta_lock, flags_spin);
	}
	*sta_id_r = sta_id;
	return ret;
}

/**
 * iwl_sta_ucode_deactivate - deactivate ucode status for a station
 *
 * priv-&gt;shrd-&gt;sta_lock must be held
 */
static void iwl_sta_ucode_deactivate(struct iwl_priv *priv, u8 sta_id)
{
	/* Ucode must be active and driver must be non active */
	if ((priv-&gt;stations[sta_id].used &amp;
	     (IWL_STA_UCODE_ACTIVE | IWL_STA_DRIVER_ACTIVE)) !=
	      IWL_STA_UCODE_ACTIVE)
		IWL_ERR(priv, "removed non active STA %u\n", sta_id);

	priv-&gt;stations[sta_id].used &amp;= ~IWL_STA_UCODE_ACTIVE;

	memset(&amp;priv-&gt;stations[sta_id], 0, sizeof(struct iwl_station_entry));
	IWL_DEBUG_ASSOC(priv, "Removed STA %u\n", sta_id);
}

static int iwl_send_remove_station(struct iwl_priv *priv,
				   const u8 *addr, int sta_id,
				   bool temporary)
{
	struct iwl_rx_packet *pkt;
	int ret;

	unsigned long flags_spin;
	struct iwl_rem_sta_cmd rm_sta_cmd;

	struct iwl_host_cmd cmd = {
		.id = REPLY_REMOVE_STA,
		.len = { sizeof(struct iwl_rem_sta_cmd), },
		.flags = CMD_SYNC,
		.data = { &amp;rm_sta_cmd, },
	};

	memset(&amp;rm_sta_cmd, 0, sizeof(rm_sta_cmd));
	rm_sta_cmd.num_sta = 1;
	memcpy(&amp;rm_sta_cmd.addr, addr, ETH_ALEN);

	cmd.flags |= CMD_WANT_SKB;

	ret = iwl_trans_send_cmd(trans(priv), &amp;cmd);

	if (ret)
		return ret;

	pkt = (struct iwl_rx_packet *)cmd.reply_page;
	if (pkt-&gt;hdr.flags &amp; IWL_CMD_FAILED_MSK) {
		IWL_ERR(priv, "Bad return from REPLY_REMOVE_STA (0x%08X)\n",
			  pkt-&gt;hdr.flags);
		ret = -EIO;
	}

	if (!ret) {
		switch (pkt-&gt;u.rem_sta.status) {
		case REM_STA_SUCCESS_MSK:
			if (!temporary) {
				spin_lock_irqsave(&amp;priv-&gt;shrd-&gt;sta_lock,
					flags_spin);
				iwl_sta_ucode_deactivate(priv, sta_id);
				spin_unlock_irqrestore(&amp;priv-&gt;shrd-&gt;sta_lock,
					flags_spin);
			}
			IWL_DEBUG_ASSOC(priv, "REPLY_REMOVE_STA PASSED\n");
			break;
		default:
			ret = -EIO;
			IWL_ERR(priv, "REPLY_REMOVE_STA failed\n");
			break;
		}
	}
	iwl_free_pages(priv-&gt;shrd, cmd.reply_page);

	return ret;
}

/**
 * iwl_remove_station - Remove driver's knowledge of station.
 */
int iwl_remove_station(struct iwl_priv *priv, const u8 sta_id,
		       const u8 *addr)
{
	unsigned long flags;
	u8 tid;

	if (!iwl_is_ready(priv-&gt;shrd)) {
		IWL_DEBUG_INFO(priv,
			"Unable to remove station %pM, device not ready.\n",
			addr);
		/*
		 * It is typical for stations to be removed when we are
		 * going down. Return success since device will be down
		 * soon anyway
		 */
		return 0;
	}

	IWL_DEBUG_ASSOC(priv, "Removing STA from driver:%d  %pM\n",
			sta_id, addr);

	if (WARN_ON(sta_id == IWL_INVALID_STATION))
		return -EINVAL;

	spin_lock_irqsave(&amp;priv-&gt;shrd-&gt;sta_lock, flags);

	if (!(priv-&gt;stations[sta_id].used &amp; IWL_STA_DRIVER_ACTIVE)) {
		IWL_DEBUG_INFO(priv, "Removing %pM but non DRIVER active\n",
				addr);
		goto out_err;
	}

	if (!(priv-&gt;stations[sta_id].used &amp; IWL_STA_UCODE_ACTIVE)) {
		IWL_DEBUG_INFO(priv, "Removing %pM but non UCODE active\n",
				addr);
		goto out_err;
	}

	if (priv-&gt;stations[sta_id].used &amp; IWL_STA_LOCAL) {
		kfree(priv-&gt;stations[sta_id].lq);
		priv-&gt;stations[sta_id].lq = NULL;
	}

	for (tid = 0; tid &lt; IWL_MAX_TID_COUNT; tid++)
		memset(&amp;priv-&gt;tid_data[sta_id][tid], 0,
			sizeof(priv-&gt;tid_data[sta_id][tid]));

	priv-&gt;stations[sta_id].used &amp;= ~IWL_STA_DRIVER_ACTIVE;

	priv-&gt;num_stations--;

	if (WARN_ON(priv-&gt;num_stations &lt; 0))
		priv-&gt;num_stations = 0;

	spin_unlock_irqrestore(&amp;priv-&gt;shrd-&gt;sta_lock, flags);

	return iwl_send_remove_station(priv, addr, sta_id, false);
out_err:
	spin_unlock_irqrestore(&amp;priv-&gt;shrd-&gt;sta_lock, flags);
	return -EINVAL;
}

/**
 * iwl_clear_ucode_stations - clear ucode station table bits
 *
 * This function clears all the bits in the driver indicating
 * which stations are active in the ucode. Call when something
 * other than explicit station management would cause this in
 * the ucode, e.g. unassociated RXON.
 */
void iwl_clear_ucode_stations(struct iwl_priv *priv,
			      struct iwl_rxon_context *ctx)
{
	int i;
	unsigned long flags_spin;
	bool cleared = false;

	IWL_DEBUG_INFO(priv, "Clearing ucode stations in driver\n");

	spin_lock_irqsave(&amp;priv-&gt;shrd-&gt;sta_lock, flags_spin);
	for (i = 0; i &lt; IWLAGN_STATION_COUNT; i++) {
		if (ctx &amp;&amp; ctx-&gt;ctxid != priv-&gt;stations[i].ctxid)
			continue;

		if (priv-&gt;stations[i].used &amp; IWL_STA_UCODE_ACTIVE) {
			IWL_DEBUG_INFO(priv,
				"Clearing ucode active for station %d\n", i);
			priv-&gt;stations[i].used &amp;= ~IWL_STA_UCODE_ACTIVE;
			cleared = true;
		}
	}
	spin_unlock_irqrestore(&amp;priv-&gt;shrd-&gt;sta_lock, flags_spin);

	if (!cleared)
		IWL_DEBUG_INFO(priv,
			       "No active stations found to be cleared\n");
}

/**
 * iwl_restore_stations() - Restore driver known stations to device
 *
 * All stations considered active by driver, but not present in ucode, is
 * restored.
 *
 * Function sleeps.
 */
void iwl_restore_stations(struct iwl_priv *priv, struct iwl_rxon_context *ctx)
{
	struct iwl_addsta_cmd sta_cmd;
	struct iwl_link_quality_cmd lq;
	unsigned long flags_spin;
	int i;
	bool found = false;
	int ret;
	bool send_lq;

	if (!iwl_is_ready(priv-&gt;shrd)) {
		IWL_DEBUG_INFO(priv,
			       "Not ready yet, not restoring any stations.\n");
		return;
	}

	IWL_DEBUG_ASSOC(priv, "Restoring all known stations ... start.\n");
	spin_lock_irqsave(&amp;priv-&gt;shrd-&gt;sta_lock, flags_spin);
	for (i = 0; i &lt; IWLAGN_STATION_COUNT; i++) {
		if (ctx-&gt;ctxid != priv-&gt;stations[i].ctxid)
			continue;
		if ((priv-&gt;stations[i].used &amp; IWL_STA_DRIVER_ACTIVE) &amp;&amp;
			    !(priv-&gt;stations[i].used &amp; IWL_STA_UCODE_ACTIVE)) {
			IWL_DEBUG_ASSOC(priv, "Restoring sta %pM\n",
					priv-&gt;stations[i].sta.sta.addr);
			priv-&gt;stations[i].sta.mode = 0;
			priv-&gt;stations[i].used |= IWL_STA_UCODE_INPROGRESS;
			found = true;
		}
	}

	for (i = 0; i &lt; IWLAGN_STATION_COUNT; i++) {
		if ((priv-&gt;stations[i].used &amp; IWL_STA_UCODE_INPROGRESS)) {
			memcpy(&amp;sta_cmd, &amp;priv-&gt;stations[i].sta,
			       sizeof(struct iwl_addsta_cmd));
			send_lq = false;
			if (priv-&gt;stations[i].lq) {
				if (priv-&gt;shrd-&gt;wowlan)
					iwl_sta_fill_lq(priv, ctx, i, &amp;lq);
				else
					memcpy(&amp;lq, priv-&gt;stations[i].lq,
					       sizeof(struct iwl_link_quality_cmd));
				send_lq = true;
			}
			spin_unlock_irqrestore(&amp;priv-&gt;shrd-&gt;sta_lock,
					       flags_spin);
			ret = iwl_send_add_sta(priv, &amp;sta_cmd, CMD_SYNC);
			if (ret) {
				spin_lock_irqsave(&amp;priv-&gt;shrd-&gt;sta_lock,
						  flags_spin);
				IWL_ERR(priv, "Adding station %pM failed.\n",
					priv-&gt;stations[i].sta.sta.addr);
				priv-&gt;stations[i].used &amp;=
						~IWL_STA_DRIVER_ACTIVE;
				priv-&gt;stations[i].used &amp;=
						~IWL_STA_UCODE_INPROGRESS;
				spin_unlock_irqrestore(&amp;priv-&gt;shrd-&gt;sta_lock,
						       flags_spin);
			}
			/*
			 * Rate scaling has already been initialized, send
			 * current LQ command
			 */
			if (send_lq)
				iwl_send_lq_cmd(priv, ctx, &amp;lq,
						CMD_SYNC, true);
			spin_lock_irqsave(&amp;priv-&gt;shrd-&gt;sta_lock, flags_spin);
			priv-&gt;stations[i].used &amp;= ~IWL_STA_UCODE_INPROGRESS;
		}
	}

	spin_unlock_irqrestore(&amp;priv-&gt;shrd-&gt;sta_lock, flags_spin);
	if (!found)
		IWL_DEBUG_INFO(priv, "Restoring all known stations .... "
			"no stations to be restored.\n");
	else
		IWL_DEBUG_INFO(priv, "Restoring all known stations .... "
			"complete.\n");
}

void iwl_reprogram_ap_sta(struct iwl_priv *priv, struct iwl_rxon_context *ctx)
{
	unsigned long flags;
	int sta_id = ctx-&gt;ap_sta_id;
	int ret;
	struct iwl_addsta_cmd sta_cmd;
	struct iwl_link_quality_cmd lq;
	bool active, have_lq = false;

	spin_lock_irqsave(&amp;priv-&gt;shrd-&gt;sta_lock, flags);
	if (!(priv-&gt;stations[sta_id].used &amp; IWL_STA_DRIVER_ACTIVE)) {
		spin_unlock_irqrestore(&amp;priv-&gt;shrd-&gt;sta_lock, flags);
		return;
	}

	memcpy(&amp;sta_cmd, &amp;priv-&gt;stations[sta_id].sta, sizeof(sta_cmd));
	sta_cmd.mode = 0;
	if (priv-&gt;stations[sta_id].lq) {
		memcpy(&amp;lq, priv-&gt;stations[sta_id].lq, sizeof(lq));
		have_lq = true;
	}

	active = priv-&gt;stations[sta_id].used &amp; IWL_STA_UCODE_ACTIVE;
	priv-&gt;stations[sta_id].used &amp;= ~IWL_STA_DRIVER_ACTIVE;
	spin_unlock_irqrestore(&amp;priv-&gt;shrd-&gt;sta_lock, flags);

	if (active) {
		ret = iwl_send_remove_station(
			priv, priv-&gt;stations[sta_id].sta.sta.addr,
			sta_id, true);
		if (ret)
			IWL_ERR(priv, "failed to remove STA %pM (%d)\n",
				priv-&gt;stations[sta_id].sta.sta.addr, ret);
	}
	spin_lock_irqsave(&amp;priv-&gt;shrd-&gt;sta_lock, flags);
	priv-&gt;stations[sta_id].used |= IWL_STA_DRIVER_ACTIVE;
	spin_unlock_irqrestore(&amp;priv-&gt;shrd-&gt;sta_lock, flags);

	ret = iwl_send_add_sta(priv, &amp;sta_cmd, CMD_SYNC);
	if (ret)
		IWL_ERR(priv, "failed to re-add STA %pM (%d)\n",
			priv-&gt;stations[sta_id].sta.sta.addr, ret);
	if (have_lq)
		iwl_send_lq_cmd(priv, ctx, &amp;lq, CMD_SYNC, true);
}

int iwl_get_free_ucode_key_offset(struct iwl_priv *priv)
{
	int i;

	for (i = 0; i &lt; priv-&gt;sta_key_max_num; i++)
		if (!test_and_set_bit(i, &amp;priv-&gt;ucode_key_table))
			return i;

	return WEP_INVALID_OFFSET;
}

void iwl_dealloc_bcast_stations(struct iwl_priv *priv)
{
	unsigned long flags;
	int i;

	spin_lock_irqsave(&amp;priv-&gt;shrd-&gt;sta_lock, flags);
	for (i = 0; i &lt; IWLAGN_STATION_COUNT; i++) {
		if (!(priv-&gt;stations[i].used &amp; IWL_STA_BCAST))
			continue;

		priv-&gt;stations[i].used &amp;= ~IWL_STA_UCODE_ACTIVE;
		priv-&gt;num_stations--;
		if (WARN_ON(priv-&gt;num_stations &lt; 0))
			priv-&gt;num_stations = 0;
		kfree(priv-&gt;stations[i].lq);
		priv-&gt;stations[i].lq = NULL;
	}
	spin_unlock_irqrestore(&amp;priv-&gt;shrd-&gt;sta_lock, flags);
}

#ifdef CONFIG_IWLWIFI_DEBUG
static void iwl_dump_lq_cmd(struct iwl_priv *priv,
			   struct iwl_link_quality_cmd *lq)
{
	int i;
	IWL_DEBUG_RATE(priv, "lq station id 0x%x\n", lq-&gt;sta_id);
	IWL_DEBUG_RATE(priv, "lq ant 0x%X 0x%X\n",
		       lq-&gt;general_params.single_stream_ant_msk,
		       lq-&gt;general_params.dual_stream_ant_msk);

	for (i = 0; i &lt; LINK_QUAL_MAX_RETRY_NUM; i++)
		IWL_DEBUG_RATE(priv, "lq index %d 0x%X\n",
			       i, lq-&gt;rs_table[i].rate_n_flags);
}
#else
static inline void iwl_dump_lq_cmd(struct iwl_priv *priv,
				   struct iwl_link_quality_cmd *lq)
{
}
#endif

/**
 * is_lq_table_valid() - Test one aspect of LQ cmd for validity
 *
 * It sometimes happens when a HT rate has been in use and we
 * loose connectivity with AP then mac80211 will first tell us that the
 * current channel is not HT anymore before removing the station. In such a
 * scenario the RXON flags will be updated to indicate we are not
 * communicating HT anymore, but the LQ command may still contain HT rates.
 * Test for this to prevent driver from sending LQ command between the time
 * RXON flags are updated and when LQ command is updated.
 */
static bool is_lq_table_valid(struct iwl_priv *priv,
			      struct iwl_rxon_context *ctx,
			      struct iwl_link_quality_cmd *lq)
{
	int i;

	if (ctx-&gt;ht.enabled)
		return true;

	IWL_DEBUG_INFO(priv, "Channel %u is not an HT channel\n",
		       ctx-&gt;active.channel);
	for (i = 0; i &lt; LINK_QUAL_MAX_RETRY_NUM; i++) {
		if (le32_to_cpu(lq-&gt;rs_table[i].rate_n_flags) &amp;
		    RATE_MCS_HT_MSK) {
			IWL_DEBUG_INFO(priv,
				       "index %d of LQ expects HT channel\n",
				       i);
			return false;
		}
	}
	return true;
}

/**
 * iwl_send_lq_cmd() - Send link quality command
 * @init: This command is sent as part of station initialization right
 *        after station has been added.
 *
 * The link quality command is sent as the last step of station creation.
 * This is the special case in which init is set and we call a callback in
 * this case to clear the state indicating that station creation is in
 * progress.
 */
int iwl_send_lq_cmd(struct iwl_priv *priv, struct iwl_rxon_context *ctx,
		    struct iwl_link_quality_cmd *lq, u8 flags, bool init)
{
	int ret = 0;
	unsigned long flags_spin;

	struct iwl_host_cmd cmd = {
		.id = REPLY_TX_LINK_QUALITY_CMD,
		.len = { sizeof(struct iwl_link_quality_cmd), },
		.flags = flags,
		.data = { lq, },
	};

	if (WARN_ON(lq-&gt;sta_id == IWL_INVALID_STATION))
		return -EINVAL;


	spin_lock_irqsave(&amp;priv-&gt;shrd-&gt;sta_lock, flags_spin);
	if (!(priv-&gt;stations[lq-&gt;sta_id].used &amp; IWL_STA_DRIVER_ACTIVE)) {
		spin_unlock_irqrestore(&amp;priv-&gt;shrd-&gt;sta_lock, flags_spin);
		return -EINVAL;
	}
	spin_unlock_irqrestore(&amp;priv-&gt;shrd-&gt;sta_lock, flags_spin);

	iwl_dump_lq_cmd(priv, lq);
	if (WARN_ON(init &amp;&amp; (cmd.flags &amp; CMD_ASYNC)))
		return -EINVAL;

	if (is_lq_table_valid(priv, ctx, lq))
		ret = iwl_trans_send_cmd(trans(priv), &amp;cmd);
	else
		ret = -EINVAL;

	if (cmd.flags &amp; CMD_ASYNC)
		return ret;

	if (init) {
		IWL_DEBUG_INFO(priv, "init LQ command complete, "
			       "clearing sta addition status for sta %d\n",
			       lq-&gt;sta_id);
		spin_lock_irqsave(&amp;priv-&gt;shrd-&gt;sta_lock, flags_spin);
		priv-&gt;stations[lq-&gt;sta_id].used &amp;= ~IWL_STA_UCODE_INPROGRESS;
		spin_unlock_irqrestore(&amp;priv-&gt;shrd-&gt;sta_lock, flags_spin);
	}
	return ret;
}


void iwl_sta_fill_lq(struct iwl_priv *priv, struct iwl_rxon_context *ctx,
		     u8 sta_id, struct iwl_link_quality_cmd *link_cmd)
{
	int i, r;
	u32 rate_flags = 0;
	__le32 rate_n_flags;

	lockdep_assert_held(&amp;priv-&gt;shrd-&gt;mutex);

	memset(link_cmd, 0, sizeof(*link_cmd));

	/* Set up the rate scaling to start at selected rate, fall back
	 * all the way down to 1M in IEEE order, and then spin on 1M */
	if (priv-&gt;band == IEEE80211_BAND_5GHZ)
		r = IWL_RATE_6M_INDEX;
	else if (ctx &amp;&amp; ctx-&gt;vif &amp;&amp; ctx-&gt;vif-&gt;p2p)
		r = IWL_RATE_6M_INDEX;
	else
		r = IWL_RATE_1M_INDEX;

	if (r &gt;= IWL_FIRST_CCK_RATE &amp;&amp; r &lt;= IWL_LAST_CCK_RATE)
		rate_flags |= RATE_MCS_CCK_MSK;

	rate_flags |= first_antenna(hw_params(priv).valid_tx_ant) &lt;&lt;
				RATE_MCS_ANT_POS;
	rate_n_flags = iwl_hw_set_rate_n_flags(iwl_rates[r].plcp, rate_flags);
	for (i = 0; i &lt; LINK_QUAL_MAX_RETRY_NUM; i++)
		link_cmd-&gt;rs_table[i].rate_n_flags = rate_n_flags;

	link_cmd-&gt;general_params.single_stream_ant_msk =
			first_antenna(hw_params(priv).valid_tx_ant);

	link_cmd-&gt;general_params.dual_stream_ant_msk =
		hw_params(priv).valid_tx_ant &amp;
		~first_antenna(hw_params(priv).valid_tx_ant);
	if (!link_cmd-&gt;general_params.dual_stream_ant_msk) {
		link_cmd-&gt;general_params.dual_stream_ant_msk = ANT_AB;
	} else if (num_of_ant(hw_params(priv).valid_tx_ant) == 2) {
		link_cmd-&gt;general_params.dual_stream_ant_msk =
			hw_params(priv).valid_tx_ant;
	}

	link_cmd-&gt;agg_params.agg_dis_start_th =
		LINK_QUAL_AGG_DISABLE_START_DEF;
	link_cmd-&gt;agg_params.agg_time_limit =
		cpu_to_le16(LINK_QUAL_AGG_TIME_LIMIT_DEF);

	link_cmd-&gt;sta_id = sta_id;
}

static struct iwl_link_quality_cmd *
iwl_sta_alloc_lq(struct iwl_priv *priv, struct iwl_rxon_context *ctx,
		 u8 sta_id)
{
	struct iwl_link_quality_cmd *link_cmd;

	link_cmd = kzalloc(sizeof(struct iwl_link_quality_cmd), GFP_KERNEL);
	if (!link_cmd) {
		IWL_ERR(priv, "Unable to allocate memory for LQ cmd.\n");
		return NULL;
	}

	iwl_sta_fill_lq(priv, ctx, sta_id, link_cmd);

	return link_cmd;
}

/*
 * iwlagn_add_bssid_station - Add the special IBSS BSSID station
 *
 * Function sleeps.
 */
int iwlagn_add_bssid_station(struct iwl_priv *priv,
			     struct iwl_rxon_context *ctx,
			     const u8 *addr, u8 *sta_id_r)
{
	int ret;
	u8 sta_id;
	struct iwl_link_quality_cmd *link_cmd;
	unsigned long flags;

	if (sta_id_r)
		*sta_id_r = IWL_INVALID_STATION;

	ret = iwl_add_station_common(priv, ctx, addr, 0, NULL, &amp;sta_id);
	if (ret) {
		IWL_ERR(priv, "Unable to add station %pM\n", addr);
		return ret;
	}

	if (sta_id_r)
		*sta_id_r = sta_id;

	spin_lock_irqsave(&amp;priv-&gt;shrd-&gt;sta_lock, flags);
	priv-&gt;stations[sta_id].used |= IWL_STA_LOCAL;
	spin_unlock_irqrestore(&amp;priv-&gt;shrd-&gt;sta_lock, flags);

	/* Set up default rate scaling table in device's station table */
	link_cmd = iwl_sta_alloc_lq(priv, ctx, sta_id);
	if (!link_cmd) {
		IWL_ERR(priv,
			"Unable to initialize rate scaling for station %pM.\n",
			addr);
		return -ENOMEM;
	}

	ret = iwl_send_lq_cmd(priv, ctx, link_cmd, CMD_SYNC, true);
	if (ret)
		IWL_ERR(priv, "Link quality command failed (%d)\n", ret);

	spin_lock_irqsave(&amp;priv-&gt;shrd-&gt;sta_lock, flags);
	priv-&gt;stations[sta_id].lq = link_cmd;
	spin_unlock_irqrestore(&amp;priv-&gt;shrd-&gt;sta_lock, flags);

	return 0;
}

/*
 * static WEP keys
 *
 * For each context, the device has a table of 4 static WEP keys
 * (one for each key index) that is updated with the following
 * commands.
 */

static int iwl_send_static_wepkey_cmd(struct iwl_priv *priv,
				      struct iwl_rxon_context *ctx,
				      bool send_if_empty)
{
	int i, not_empty = 0;
	u8 buff[sizeof(struct iwl_wep_cmd) +
		sizeof(struct iwl_wep_key) * WEP_KEYS_MAX];
	struct iwl_wep_cmd *wep_cmd = (struct iwl_wep_cmd *)buff;
	size_t cmd_size  = sizeof(struct iwl_wep_cmd);
	struct iwl_host_cmd cmd = {
		.id = ctx-&gt;wep_key_cmd,
		.data = { wep_cmd, },
		.flags = CMD_SYNC,
	};

	might_sleep();

	memset(wep_cmd, 0, cmd_size +
			(sizeof(struct iwl_wep_key) * WEP_KEYS_MAX));

	for (i = 0; i &lt; WEP_KEYS_MAX ; i++) {
		wep_cmd-&gt;key[i].key_index = i;
		if (ctx-&gt;wep_keys[i].key_size) {
			wep_cmd-&gt;key[i].key_offset = i;
			not_empty = 1;
		} else {
			wep_cmd-&gt;key[i].key_offset = WEP_INVALID_OFFSET;
		}

		wep_cmd-&gt;key[i].key_size = ctx-&gt;wep_keys[i].key_size;
		memcpy(&amp;wep_cmd-&gt;key[i].key[3], ctx-&gt;wep_keys[i].key,
				ctx-&gt;wep_keys[i].key_size);
	}

	wep_cmd-&gt;global_key_type = WEP_KEY_WEP_TYPE;
	wep_cmd-&gt;num_keys = WEP_KEYS_MAX;

	cmd_size += sizeof(struct iwl_wep_key) * WEP_KEYS_MAX;

	cmd.len[0] = cmd_size;

	if (not_empty || send_if_empty)
		return iwl_trans_send_cmd(trans(priv), &amp;cmd);
	else
		return 0;
}

int iwl_restore_default_wep_keys(struct iwl_priv *priv,
				 struct iwl_rxon_context *ctx)
{
	lockdep_assert_held(&amp;priv-&gt;shrd-&gt;mutex);

	return iwl_send_static_wepkey_cmd(priv, ctx, false);
}

int iwl_remove_default_wep_key(struct iwl_priv *priv,
			       struct iwl_rxon_context *ctx,
			       struct ieee80211_key_conf *keyconf)
{
	int ret;

	lockdep_assert_held(&amp;priv-&gt;shrd-&gt;mutex);

	IWL_DEBUG_WEP(priv, "Removing default WEP key: idx=%d\n",
		      keyconf-&gt;keyidx);

	memset(&amp;ctx-&gt;wep_keys[keyconf-&gt;keyidx], 0, sizeof(ctx-&gt;wep_keys[0]));
	if (iwl_is_rfkill(priv-&gt;shrd)) {
		IWL_DEBUG_WEP(priv,
			"Not sending REPLY_WEPKEY command due to RFKILL.\n");
		/* but keys in device are clear anyway so return success */
		return 0;
	}
	ret = iwl_send_static_wepkey_cmd(priv, ctx, 1);
	IWL_DEBUG_WEP(priv, "Remove default WEP key: idx=%d ret=%d\n",
		      keyconf-&gt;keyidx, ret);

	return ret;
}

int iwl_set_default_wep_key(struct iwl_priv *priv,
			    struct iwl_rxon_context *ctx,
			    struct ieee80211_key_conf *keyconf)
{
	int ret;

	lockdep_assert_held(&amp;priv-&gt;shrd-&gt;mutex);

	if (keyconf-&gt;keylen != WEP_KEY_LEN_128 &amp;&amp;
	    keyconf-&gt;keylen != WEP_KEY_LEN_64) {
		IWL_DEBUG_WEP(priv,
			      "Bad WEP key length %d\n", keyconf-&gt;keylen);
		return -EINVAL;
	}

	keyconf-&gt;hw_key_idx = IWLAGN_HW_KEY_DEFAULT;

	ctx-&gt;wep_keys[keyconf-&gt;keyidx].key_size = keyconf-&gt;keylen;
	memcpy(&amp;ctx-&gt;wep_keys[keyconf-&gt;keyidx].key, &amp;keyconf-&gt;key,
							keyconf-&gt;keylen);

	ret = iwl_send_static_wepkey_cmd(priv, ctx, false);
	IWL_DEBUG_WEP(priv, "Set default WEP key: len=%d idx=%d ret=%d\n",
		keyconf-&gt;keylen, keyconf-&gt;keyidx, ret);

	return ret;
}

/*
 * dynamic (per-station) keys
 *
 * The dynamic keys are a little more complicated. The device has
 * a key cache of up to STA_KEY_MAX_NUM/STA_KEY_MAX_NUM_PAN keys.
 * These are linked to stations by a table that contains an index
 * into the key table for each station/key index/{mcast,unicast},
 * i.e. it's basically an array of pointers like this:
 *	key_offset_t key_mapping[NUM_STATIONS][4][2];
 * (it really works differently, but you can think of it as such)
 *
 * The key uploading and linking happens in the same command, the
 * add station command with STA_MODIFY_KEY_MASK.
 */

static u8 iwlagn_key_sta_id(struct iwl_priv *priv,
			    struct ieee80211_vif *vif,
			    struct ieee80211_sta *sta)
{
	struct iwl_vif_priv *vif_priv = (void *)vif-&gt;drv_priv;
	u8 sta_id = IWL_INVALID_STATION;

	if (sta)
		sta_id = iwl_sta_id(sta);

	/*
	 * The device expects GTKs for station interfaces to be
	 * installed as GTKs for the AP station. If we have no
	 * station ID, then use the ap_sta_id in that case.
	 */
	if (!sta &amp;&amp; vif &amp;&amp; vif_priv-&gt;ctx) {
		switch (vif-&gt;type) {
		case NL80211_IFTYPE_STATION:
			sta_id = vif_priv-&gt;ctx-&gt;ap_sta_id;
			break;
		default:
			/*
			 * In all other cases, the key will be
			 * used either for TX only or is bound
			 * to a station already.
			 */
			break;
		}
	}

	return sta_id;
}

static int iwlagn_send_sta_key(struct iwl_priv *priv,
			       struct ieee80211_key_conf *keyconf,
			       u8 sta_id, u32 tkip_iv32, u16 *tkip_p1k,
			       u32 cmd_flags)
{
	unsigned long flags;
	__le16 key_flags;
	struct iwl_addsta_cmd sta_cmd;
	int i;

	spin_lock_irqsave(&amp;priv-&gt;shrd-&gt;sta_lock, flags);
	memcpy(&amp;sta_cmd, &amp;priv-&gt;stations[sta_id].sta, sizeof(sta_cmd));
	spin_unlock_irqrestore(&amp;priv-&gt;shrd-&gt;sta_lock, flags);

	key_flags = cpu_to_le16(keyconf-&gt;keyidx &lt;&lt; STA_KEY_FLG_KEYID_POS);
	key_flags |= STA_KEY_FLG_MAP_KEY_MSK;

	switch (keyconf-&gt;cipher) {
	case WLAN_CIPHER_SUITE_CCMP:
		key_flags |= STA_KEY_FLG_CCMP;
		memcpy(sta_cmd.key.key, keyconf-&gt;key, keyconf-&gt;keylen);
		break;
	case WLAN_CIPHER_SUITE_TKIP:
		key_flags |= STA_KEY_FLG_TKIP;
		sta_cmd.key.tkip_rx_tsc_byte2 = tkip_iv32;
		for (i = 0; i &lt; 5; i++)
			sta_cmd.key.tkip_rx_ttak[i] = cpu_to_le16(tkip_p1k[i]);
		memcpy(sta_cmd.key.key, keyconf-&gt;key, keyconf-&gt;keylen);
		break;
	case WLAN_CIPHER_SUITE_WEP104:
		key_flags |= STA_KEY_FLG_KEY_SIZE_MSK;
		/* fall through */
	case WLAN_CIPHER_SUITE_WEP40:
		key_flags |= STA_KEY_FLG_WEP;
		memcpy(&amp;sta_cmd.key.key[3], keyconf-&gt;key, keyconf-&gt;keylen);
		break;
	default:
		WARN_ON(1);
		return -EINVAL;
	}

	if (!(keyconf-&gt;flags &amp; IEEE80211_KEY_FLAG_PAIRWISE))
		key_flags |= STA_KEY_MULTICAST_MSK;

	/* key pointer (offset) */
	sta_cmd.key.key_offset = keyconf-&gt;hw_key_idx;

	sta_cmd.key.key_flags = key_flags;
	sta_cmd.mode = STA_CONTROL_MODIFY_MSK;
	sta_cmd.sta.modify_mask = STA_MODIFY_KEY_MASK;

	return iwl_send_add_sta(priv, &amp;sta_cmd, cmd_flags);
}

void iwl_update_tkip_key(struct iwl_priv *priv,
			 struct ieee80211_vif *vif,
			 struct ieee80211_key_conf *keyconf,
			 struct ieee80211_sta *sta, u32 iv32, u16 *phase1key)
{
	u8 sta_id = iwlagn_key_sta_id(priv, vif, sta);

	if (sta_id == IWL_INVALID_STATION)
		return;

	if (iwl_scan_cancel(priv)) {
		/* cancel scan failed, just live w/ bad key and rely
		   briefly on SW decryption */
		return;
	}

	iwlagn_send_sta_key(priv, keyconf, sta_id,
			    iv32, phase1key, CMD_ASYNC);
}

int iwl_remove_dynamic_key(struct iwl_priv *priv,
			   struct iwl_rxon_context *ctx,
			   struct ieee80211_key_conf *keyconf,
			   struct ieee80211_sta *sta)
{
	unsigned long flags;
	struct iwl_addsta_cmd sta_cmd;
	u8 sta_id = iwlagn_key_sta_id(priv, ctx-&gt;vif, sta);

	/* if station isn't there, neither is the key */
	if (sta_id == IWL_INVALID_STATION)
		return -ENOENT;

	spin_lock_irqsave(&amp;priv-&gt;shrd-&gt;sta_lock, flags);
	memcpy(&amp;sta_cmd, &amp;priv-&gt;stations[sta_id].sta, sizeof(sta_cmd));
	if (!(priv-&gt;stations[sta_id].used &amp; IWL_STA_UCODE_ACTIVE))
		sta_id = IWL_INVALID_STATION;
	spin_unlock_irqrestore(&amp;priv-&gt;shrd-&gt;sta_lock, flags);

	if (sta_id == IWL_INVALID_STATION)
		return 0;

	lockdep_assert_held(&amp;priv-&gt;shrd-&gt;mutex);

	ctx-&gt;key_mapping_keys--;

	IWL_DEBUG_WEP(priv, "Remove dynamic key: idx=%d sta=%d\n",
		      keyconf-&gt;keyidx, sta_id);

	if (!test_and_clear_bit(keyconf-&gt;hw_key_idx, &amp;priv-&gt;ucode_key_table))
		IWL_ERR(priv, "offset %d not used in uCode key table.\n",
			keyconf-&gt;hw_key_idx);

	sta_cmd.key.key_flags = STA_KEY_FLG_NO_ENC | STA_KEY_FLG_INVALID;
	sta_cmd.key.key_offset = WEP_INVALID_OFFSET;
	sta_cmd.sta.modify_mask = STA_MODIFY_KEY_MASK;
	sta_cmd.mode = STA_CONTROL_MODIFY_MSK;

	return iwl_send_add_sta(priv, &amp;sta_cmd, CMD_SYNC);
}

int iwl_set_dynamic_key(struct iwl_priv *priv,
			struct iwl_rxon_context *ctx,
			struct ieee80211_key_conf *keyconf,
			struct ieee80211_sta *sta)
{
	struct ieee80211_key_seq seq;
	u16 p1k[5];
	int ret;
	u8 sta_id = iwlagn_key_sta_id(priv, ctx-&gt;vif, sta);
	const u8 *addr;

	if (sta_id == IWL_INVALID_STATION)
		return -EINVAL;

	lockdep_assert_held(&amp;priv-&gt;shrd-&gt;mutex);

	keyconf-&gt;hw_key_idx = iwl_get_free_ucode_key_offset(priv);
	if (keyconf-&gt;hw_key_idx == WEP_INVALID_OFFSET)
		return -ENOSPC;

	ctx-&gt;key_mapping_keys++;

	switch (keyconf-&gt;cipher) {
	case WLAN_CIPHER_SUITE_TKIP:
		if (sta)
			addr = sta-&gt;addr;
		else /* station mode case only */
			addr = ctx-&gt;active.bssid_addr;

		/* pre-fill phase 1 key into device cache */
		ieee80211_get_key_rx_seq(keyconf, 0, &amp;seq);
		ieee80211_get_tkip_rx_p1k(keyconf, addr, seq.tkip.iv32, p1k);
		ret = iwlagn_send_sta_key(priv, keyconf, sta_id,
					  seq.tkip.iv32, p1k, CMD_SYNC);
		break;
	case WLAN_CIPHER_SUITE_CCMP:
	case WLAN_CIPHER_SUITE_WEP40:
	case WLAN_CIPHER_SUITE_WEP104:
		ret = iwlagn_send_sta_key(priv, keyconf, sta_id,
					  0, NULL, CMD_SYNC);
		break;
	default:
		IWL_ERR(priv, "Unknown cipher %x\n", keyconf-&gt;cipher);
		ret = -EINVAL;
	}

	if (ret) {
		ctx-&gt;key_mapping_keys--;
		clear_bit(keyconf-&gt;hw_key_idx, &amp;priv-&gt;ucode_key_table);
	}

	IWL_DEBUG_WEP(priv, "Set dynamic key: cipher=%x len=%d idx=%d sta=%pM ret=%d\n",
		      keyconf-&gt;cipher, keyconf-&gt;keylen, keyconf-&gt;keyidx,
		      sta ? sta-&gt;addr : NULL, ret);

	return ret;
}

/**
 * iwlagn_alloc_bcast_station - add broadcast station into driver's station table.
 *
 * This adds the broadcast station into the driver's station table
 * and marks it driver active, so that it will be restored to the
 * device at the next best time.
 */
int iwlagn_alloc_bcast_station(struct iwl_priv *priv,
			       struct iwl_rxon_context *ctx)
{
	struct iwl_link_quality_cmd *link_cmd;
	unsigned long flags;
	u8 sta_id;

	spin_lock_irqsave(&amp;priv-&gt;shrd-&gt;sta_lock, flags);
	sta_id = iwl_prep_station(priv, ctx, iwl_bcast_addr, false, NULL);
	if (sta_id == IWL_INVALID_STATION) {
		IWL_ERR(priv, "Unable to prepare broadcast station\n");
		spin_unlock_irqrestore(&amp;priv-&gt;shrd-&gt;sta_lock, flags);

		return -EINVAL;
	}

	priv-&gt;stations[sta_id].used |= IWL_STA_DRIVER_ACTIVE;
	priv-&gt;stations[sta_id].used |= IWL_STA_BCAST;
	spin_unlock_irqrestore(&amp;priv-&gt;shrd-&gt;sta_lock, flags);

	link_cmd = iwl_sta_alloc_lq(priv, ctx, sta_id);
	if (!link_cmd) {
		IWL_ERR(priv,
			"Unable to initialize rate scaling for bcast station.\n");
		return -ENOMEM;
	}

	spin_lock_irqsave(&amp;priv-&gt;shrd-&gt;sta_lock, flags);
	priv-&gt;stations[sta_id].lq = link_cmd;
	spin_unlock_irqrestore(&amp;priv-&gt;shrd-&gt;sta_lock, flags);

	return 0;
}

/**
 * iwl_update_bcast_station - update broadcast station's LQ command
 *
 * Only used by iwlagn. Placed here to have all bcast station management
 * code together.
 */
int iwl_update_bcast_station(struct iwl_priv *priv,
			     struct iwl_rxon_context *ctx)
{
	unsigned long flags;
	struct iwl_link_quality_cmd *link_cmd;
	u8 sta_id = ctx-&gt;bcast_sta_id;

	link_cmd = iwl_sta_alloc_lq(priv, ctx, sta_id);
	if (!link_cmd) {
		IWL_ERR(priv, "Unable to initialize rate scaling for bcast station.\n");
		return -ENOMEM;
	}

	spin_lock_irqsave(&amp;priv-&gt;shrd-&gt;sta_lock, flags);
	if (priv-&gt;stations[sta_id].lq)
		kfree(priv-&gt;stations[sta_id].lq);
	else
		IWL_DEBUG_INFO(priv, "Bcast station rate scaling has not been initialized yet.\n");
	priv-&gt;stations[sta_id].lq = link_cmd;
	spin_unlock_irqrestore(&amp;priv-&gt;shrd-&gt;sta_lock, flags);

	return 0;
}

int iwl_update_bcast_stations(struct iwl_priv *priv)
{
	struct iwl_rxon_context *ctx;
	int ret = 0;

	for_each_context(priv, ctx) {
		ret = iwl_update_bcast_station(priv, ctx);
		if (ret)
			break;
	}

	return ret;
}

/**
 * iwl_sta_tx_modify_enable_tid - Enable Tx for this TID in station table
 */
int iwl_sta_tx_modify_enable_tid(struct iwl_priv *priv, int sta_id, int tid)
{
	unsigned long flags;
	struct iwl_addsta_cmd sta_cmd;

	lockdep_assert_held(&amp;priv-&gt;shrd-&gt;mutex);

	/* Remove "disable" flag, to enable Tx for this TID */
	spin_lock_irqsave(&amp;priv-&gt;shrd-&gt;sta_lock, flags);
	priv-&gt;stations[sta_id].sta.sta.modify_mask = STA_MODIFY_TID_DISABLE_TX;
	priv-&gt;stations[sta_id].sta.tid_disable_tx &amp;= cpu_to_le16(~(1 &lt;&lt; tid));
	priv-&gt;stations[sta_id].sta.mode = STA_CONTROL_MODIFY_MSK;
	memcpy(&amp;sta_cmd, &amp;priv-&gt;stations[sta_id].sta, sizeof(struct iwl_addsta_cmd));
	spin_unlock_irqrestore(&amp;priv-&gt;shrd-&gt;sta_lock, flags);

	return iwl_send_add_sta(priv, &amp;sta_cmd, CMD_SYNC);
}

int iwl_sta_rx_agg_start(struct iwl_priv *priv, struct ieee80211_sta *sta,
			 int tid, u16 ssn)
{
	unsigned long flags;
	int sta_id;
	struct iwl_addsta_cmd sta_cmd;

	lockdep_assert_held(&amp;priv-&gt;shrd-&gt;mutex);

	sta_id = iwl_sta_id(sta);
	if (sta_id == IWL_INVALID_STATION)
		return -ENXIO;

	spin_lock_irqsave(&amp;priv-&gt;shrd-&gt;sta_lock, flags);
	priv-&gt;stations[sta_id].sta.station_flags_msk = 0;
	priv-&gt;stations[sta_id].sta.sta.modify_mask = STA_MODIFY_ADDBA_TID_MSK;
	priv-&gt;stations[sta_id].sta.add_immediate_ba_tid = (u8)tid;
	priv-&gt;stations[sta_id].sta.add_immediate_ba_ssn = cpu_to_le16(ssn);
	priv-&gt;stations[sta_id].sta.mode = STA_CONTROL_MODIFY_MSK;
	memcpy(&amp;sta_cmd, &amp;priv-&gt;stations[sta_id].sta, sizeof(struct iwl_addsta_cmd));
	spin_unlock_irqrestore(&amp;priv-&gt;shrd-&gt;sta_lock, flags);

	return iwl_send_add_sta(priv, &amp;sta_cmd, CMD_SYNC);
}

int iwl_sta_rx_agg_stop(struct iwl_priv *priv, struct ieee80211_sta *sta,
			int tid)
{
	unsigned long flags;
	int sta_id;
	struct iwl_addsta_cmd sta_cmd;

	lockdep_assert_held(&amp;priv-&gt;shrd-&gt;mutex);

	sta_id = iwl_sta_id(sta);
	if (sta_id == IWL_INVALID_STATION) {
		IWL_ERR(priv, "Invalid station for AGG tid %d\n", tid);
		return -ENXIO;
	}

	spin_lock_irqsave(&amp;priv-&gt;shrd-&gt;sta_lock, flags);
	priv-&gt;stations[sta_id].sta.station_flags_msk = 0;
	priv-&gt;stations[sta_id].sta.sta.modify_mask = STA_MODIFY_DELBA_TID_MSK;
	priv-&gt;stations[sta_id].sta.remove_immediate_ba_tid = (u8)tid;
	priv-&gt;stations[sta_id].sta.mode = STA_CONTROL_MODIFY_MSK;
	memcpy(&amp;sta_cmd, &amp;priv-&gt;stations[sta_id].sta, sizeof(struct iwl_addsta_cmd));
	spin_unlock_irqrestore(&amp;priv-&gt;shrd-&gt;sta_lock, flags);

	return iwl_send_add_sta(priv, &amp;sta_cmd, CMD_SYNC);
}



void iwl_sta_modify_sleep_tx_count(struct iwl_priv *priv, int sta_id, int cnt)
{
	unsigned long flags;

	spin_lock_irqsave(&amp;priv-&gt;shrd-&gt;sta_lock, flags);
	priv-&gt;stations[sta_id].sta.station_flags |= STA_FLG_PWR_SAVE_MSK;
	priv-&gt;stations[sta_id].sta.station_flags_msk = STA_FLG_PWR_SAVE_MSK;
	priv-&gt;stations[sta_id].sta.sta.modify_mask =
					STA_MODIFY_SLEEP_TX_COUNT_MSK;
	priv-&gt;stations[sta_id].sta.sleep_tx_count = cpu_to_le16(cnt);
	priv-&gt;stations[sta_id].sta.mode = STA_CONTROL_MODIFY_MSK;
	iwl_send_add_sta(priv, &amp;priv-&gt;stations[sta_id].sta, CMD_ASYNC);
	spin_unlock_irqrestore(&amp;priv-&gt;shrd-&gt;sta_lock, flags);

}
</net></ilw@linux.intel.com>